# -*- coding: utf-8 -*-
#
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/stable/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import pathlib
import sys

sys.path.insert(0, pathlib.Path(__file__).parents[1].resolve().as_posix())


# -- Project information -----------------------------------------------------


def get_latest_tag():
    """
    Helper function that returns the name of the tag if the latest commit is tagged,
    otherwise it simply returns "latest"
    """

    default = "latest"

    try:
        import git
    except ImportError:
        print("GitPython module is not installed")
        return default

    current_directory = os.getcwd()

    while True:
        try:
            repo = git.Repo(current_directory)
            latest_commit = repo.head.commit
            tags = repo.tags
            for tag in tags:
                if tag.commit == latest_commit:
                    return tag.name
            # If the latest commit has not tags, then it's the latest version
            return default
        except git.InvalidGitRepositoryError:
            # No Git repository found in the current directory
            parent_directory = os.path.dirname(current_directory)
            # Check if we have reached the root directory: if yes, break the loop and then return
            if parent_directory == current_directory:
                break
            # Move to the parent directory
            current_directory = parent_directory

    print("GitPython did not find a valid repository")
    return default


project = "b2luigi"
copyright = "2018-2023, Nils Braun, Michael Eliachevitch, The Belle II Collaboration"
author = "Nils Braun, Michael Eliachevitch, The Belle II Collaboration"

# The short X.Y version
_version = get_latest_tag()
version = _version
# The full version, including alpha/beta/rc tags
release = _version


# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx_gallery.gen_gallery",
    "sphinx.ext.viewcode",
    "sphinx_sitemap",
    "notfound.extension",
]

sphinx_gallery_conf = {
    "examples_dirs": "../examples/tutorial",
    "gallery_dirs": "starterkit",
    "filename_pattern": r"Ex\d+.*\.py",
    "ignore_pattern": r"^.*\.txt$|.*\.sh$",
    "plot_gallery": False,
    "show_memory": False,
    "min_reported_time": 0,
    "remove_config_comments": True,
    "backreferences_dir": None,
    "download_all_examples": False,
    "notebook_extensions": {},  # Don't create notebook download
    "within_subsection_order": "FileNameSortKey",
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = {".rst": "restructuredtext"}

# The master toctree document.
master_doc = "index"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path .
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

autodoc_member_order = "bysource"

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_book_theme"


# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]
html_logo = "b2luigi.png"

# Custom sidebar templates, must be a dictionary that maps document names
# to template names.
#
# The default sidebars (for documents that don't match any pattern) are
# defined by theme itself.  Builtin themes are using these templates by
# default: ``['localtoc.html', 'relations.html', 'sourcelink.html',
# 'searchbox.html']``.
#
# html_sidebars = {}

# Notfound extension options
notfound_context = {"title": "Redirecting...", "body": '<meta http-equiv="refresh" content="0; url=/" />'}

# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
html_baseurl = "https://b2luigi.belle2.org/"


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',
    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        "b2luigi.tex",
        "b2luigi Documentation",
        "Nils Braun, Michael Eliachevitch, The Belle II Collaboration",
        "manual",
    ),
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, "b2luigi", "b2luigi Documentation", [author], 1)]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "b2luigi",
        "b2luigi Documentation",
        author,
        "b2luigi",
        "One line description of project.",
        "Miscellaneous",
    ),
]

# Cheating around read the docs: we need to install our project which is not possible because we
# have no setup.py. But this should work...

if os.getenv("READTHEDOCS"):
    import subprocess

    subprocess.check_call(["pip", "install", "flit"])
    subprocess.check_call(["flit", "install", "-s"], cwd="../")
